@@ -26,7 +26,6 @@
 #include "pango-fontmap.h"
 #include "pangocoretext-private.h"
 #include "pango-impl-utils.h"
-#include "modules.h"
 
 #include <Carbon/Carbon.h>
 
@@ -107,24 +106,29 @@ static PangoCoreTextFontsetKey *pango_core_text_fontset_get_key (PangoCoreTextFo
 
 typedef struct
 {
-    float bound;
-    PangoWeight weight;
+    float ct_weight;
+    PangoWeight pango_weight;
 } PangoCTWeight;
 
-const float ct_weight_min = -1.00f;
-const float ct_weight_max = 1.00f;
+#define ct_weight_min -0.7f
+#define ct_weight_max  0.8f
 
-static const PangoCTWeight ct_weight_limits[] = {
-    { -0.70, PANGO_WEIGHT_THIN},
-    { -0.50, PANGO_WEIGHT_ULTRALIGHT },
-    { -0.35, PANGO_WEIGHT_LIGHT },
-    { -0.10, PANGO_WEIGHT_BOOK },
-    {  0.10, PANGO_WEIGHT_NORMAL },
-    {  0.24, PANGO_WEIGHT_MEDIUM },
-    {  0.36, PANGO_WEIGHT_SEMIBOLD },
-    {  0.50, PANGO_WEIGHT_BOLD },
-    {  0.75, PANGO_WEIGHT_ULTRABOLD },
-    {  1.00, PANGO_WEIGHT_HEAVY }
+/* This map is based on empirical data from analyzing a large collection of
+ * fonts and comparing the opentype value with the value that OSX returns.
+ * see: https://bugzilla.gnome.org/show_bug.cgi?id=766148
+ */
+
+static const PangoCTWeight ct_weight_map[] = {
+    { ct_weight_min, PANGO_WEIGHT_THIN },
+    { -0.5, PANGO_WEIGHT_ULTRALIGHT },
+    { -0.23, PANGO_WEIGHT_LIGHT },
+    { -0.115, PANGO_WEIGHT_SEMILIGHT },
+    {  0.00, PANGO_WEIGHT_NORMAL },
+    {  0.2, PANGO_WEIGHT_MEDIUM },
+    {  0.3, PANGO_WEIGHT_SEMIBOLD },
+    {  0.4, PANGO_WEIGHT_BOLD },
+    {  0.6, PANGO_WEIGHT_ULTRABOLD },
+    {  ct_weight_max, PANGO_WEIGHT_HEAVY }
 };
 
 static const char *
@@ -285,38 +289,45 @@ cf_font_descriptor_copy_with_traits (CTFontDescriptorRef        desc,
   return new_desc;
 }
 
+static int
+lerp(float x, float x1, float x2, int y1, int y2) {
+  float dx = x2 - x1;
+  int dy = y2 - y1;
+  return y1 + (dy*(x-x1) + dx/2) / dx;
+}
+
 static PangoWeight
 ct_font_descriptor_get_weight (CTFontDescriptorRef desc)
 {
   CFDictionaryRef dict;
   CFNumberRef cf_number;
   CGFloat value;
-  PangoWeight weight;
+  PangoWeight weight = PANGO_WEIGHT_NORMAL;
 
   dict = CTFontDescriptorCopyAttribute (desc, kCTFontTraitsAttribute);
   cf_number = (CFNumberRef)CFDictionaryGetValue (dict,
                                                  kCTFontWeightTrait);
 
-  if (CFNumberGetValue (cf_number, kCFNumberCGFloatType, &value))
+  if (cf_number != NULL && CFNumberGetValue (cf_number, kCFNumberCGFloatType, &value))
     {
-      if (value < ct_weight_min || value > ct_weight_max)
-	{
-          /* This is really an error */
-          weight = PANGO_WEIGHT_NORMAL;
-	}
-      else
-	{
-	  guint i;
-	  for (i = 0; i < G_N_ELEMENTS(ct_weight_limits); i++)
-	    if (value < ct_weight_limits[i].bound)
-	      {
-                weight = ct_weight_limits[i].weight;
-                break;
-	      }
-	}
+    if (value < ct_weight_min || value > ct_weight_max)
+      {
+        /* This is really an error */
+        weight = PANGO_WEIGHT_NORMAL;
+      }
+    else
+      {
+        guint i;
+        for (i = 1; value > ct_weight_map[i].ct_weight; ++i)
+          ;
+
+        weight = lerp(value, ct_weight_map[i-1].ct_weight, ct_weight_map[i].ct_weight,
+                      ct_weight_map[i-1].pango_weight, ct_weight_map[i].pango_weight);
+
+      }
     }
- else
-   weight = PANGO_WEIGHT_NORMAL;
+  else
+    weight = PANGO_WEIGHT_NORMAL;
 
   CFRelease (dict);
 
@@ -411,6 +422,9 @@ _pango_core_text_font_description_from_ct_font_descriptor (CTFontDescriptorRef d
   else
     pango_font_description_set_style (font_desc, PANGO_STYLE_NORMAL);
 
+  if ((font_traits & kCTFontCondensedTrait) == kCTFontCondensedTrait)
+    pango_font_description_set_stretch (font_desc, PANGO_STRETCH_CONDENSED);
+
   if (ct_font_descriptor_is_small_caps (desc))
     pango_font_description_set_variant (font_desc, PANGO_VARIANT_SMALL_CAPS);
   else
@@ -713,7 +727,6 @@ static void
 pango_core_text_family_class_init (PangoCoreTextFamilyClass *klass)
 {
   GObjectClass *object_class = (GObjectClass *)klass;
-  int i;
   PangoFontFamilyClass *pfclass = PANGO_FONT_FAMILY_CLASS(klass);
 
   object_class->finalize = pango_core_text_family_finalize;
@@ -721,9 +734,6 @@ pango_core_text_family_class_init (PangoCoreTextFamilyClass *klass)
   pfclass->list_faces = pango_core_text_family_list_faces;
   pfclass->get_name = pango_core_text_family_get_name;
   pfclass->is_monospace = pango_core_text_family_is_monospace;
-
-  for (i = 0; _pango_included_core_text_modules[i].list; i++)
-    pango_module_register (&_pango_included_core_text_modules[i]);
 }
 
 static void
@@ -796,6 +806,7 @@ get_context_matrix (PangoContext *context,
 /*
  * Helper functions for PangoCoreTextFontsetKey
  */
+static const double ppi = 72.0; /* typographic points per inch */
 
 static double
 pango_core_text_font_map_get_resolution (PangoCoreTextFontMap *fontmap,
@@ -807,7 +818,7 @@ pango_core_text_font_map_get_resolution (PangoCoreTextFontMap *fontmap,
   /* FIXME: acquire DPI from CoreText using some deafault font */
   g_warning ("FIXME: returning default DPI");
 
-  return 72.0;
+  return ppi;
 }
 
 static int
@@ -816,14 +827,16 @@ get_scaled_size (PangoCoreTextFontMap       *fontmap,
                  const PangoFontDescription *desc)
 {
   double size = pango_font_description_get_size (desc);
+  PangoMatrix *matrix = pango_context_get_matrix (context);
+  double scale_factor = pango_matrix_get_font_scale_factor (matrix);
+  
+  if (!pango_font_description_get_size_is_absolute(desc))
+  {
+    double dpi = pango_core_text_font_map_get_resolution (fontmap, context);
+    size *= (dpi/ppi);
+  }
 
-  if (!pango_font_description_get_size_is_absolute (desc))
-    {
-      double dpi = pango_core_text_font_map_get_resolution (fontmap, context);
-      size = size * dpi / 72.;
-    }
-
-  return .5 + pango_matrix_get_font_scale_factor (pango_context_get_matrix (context)) * size;
+  return .5 +  scale_factor * size;
 }
 
 
@@ -836,7 +849,7 @@ struct _PangoCoreTextFontsetKey
   PangoLanguage *language;
   PangoFontDescription *desc;
   PangoMatrix matrix;
-  int pixelsize;
+  int pointsize;
   double resolution;
   PangoGravity gravity;
   gpointer context_key;
@@ -855,7 +868,7 @@ pango_core_text_fontset_key_init (PangoCoreTextFontsetKey    *key,
   key->fontmap = fontmap;
   get_context_matrix (context, &key->matrix);
   key->language = language;
-  key->pixelsize = get_scaled_size (fontmap, context, desc);
+  key->pointsize = get_scaled_size (fontmap, context, desc);
   key->resolution = pango_core_text_font_map_get_resolution (fontmap, context);
   key->gravity = pango_context_get_gravity (context);
   key->desc = pango_font_description_copy_static (desc);
@@ -875,7 +888,7 @@ pango_core_text_fontset_key_copy (const PangoCoreTextFontsetKey *old)
   key->fontmap = old->fontmap;
   key->matrix = old->matrix;
   key->language = old->language;
-  key->pixelsize = old->pixelsize;
+  key->pointsize = old->pointsize;
   key->resolution = old->resolution;
   key->gravity = old->gravity;
   key->desc = pango_font_description_copy (old->desc);
@@ -919,7 +932,7 @@ pango_core_text_fontset_key_equal (const PangoCoreTextFontsetKey *key_a,
                                    const PangoCoreTextFontsetKey *key_b)
 {
   if (key_a->language == key_b->language &&
-      key_a->pixelsize == key_b->pixelsize &&
+      key_a->pointsize == key_b->pointsize &&
       key_a->resolution == key_b->resolution &&
       key_a->gravity == key_b->gravity &&
       pango_font_description_equal (key_a->desc, key_b->desc) &&
@@ -969,7 +982,7 @@ struct _PangoCoreTextFontKey
   CTFontDescriptorRef ctfontdescriptor;
   PangoMatrix matrix;
   PangoGravity gravity;
-  int pixelsize;
+  int pointsize;
   double resolution;
   gboolean synthetic_italic;
   gpointer context_key;
@@ -985,7 +998,7 @@ pango_core_text_font_key_init (PangoCoreTextFontKey    *key,
   key->fontmap = ctfontmap;
   key->ctfontdescriptor = ctdescriptor;
   key->matrix = *pango_core_text_fontset_key_get_matrix (fontset_key);
-  key->pixelsize = fontset_key->pixelsize;
+  key->pointsize = fontset_key->pointsize;
   key->resolution = fontset_key->resolution;
   key->synthetic_italic = synthetic_italic;
   key->gravity = pango_core_text_fontset_key_get_gravity (fontset_key);
@@ -1001,7 +1014,7 @@ pango_core_text_font_key_copy (const PangoCoreTextFontKey *old)
   key->ctfontdescriptor = old->ctfontdescriptor;
   CFRetain (key->ctfontdescriptor);
   key->matrix = old->matrix;
-  key->pixelsize = old->pixelsize;
+  key->pointsize = old->pointsize;
   key->resolution = old->resolution;
   key->synthetic_italic = old->synthetic_italic;
   key->gravity = old->gravity;
@@ -1047,7 +1060,7 @@ pango_core_text_font_key_equal (const PangoCoreTextFontKey *key_a,
   if (CFEqual (key_a->ctfontdescriptor, key_b->ctfontdescriptor) &&
       memcmp (&key_a->matrix, &key_b->matrix, 4 * sizeof (double)) == 0 &&
       key_a->gravity == key_b->gravity &&
-      key_a->pixelsize == key_b->pixelsize &&
+      key_a->pointsize == key_b->pointsize &&
       key_a->resolution == key_b->resolution &&
       key_a->synthetic_italic == key_b->synthetic_italic)
     {
@@ -1063,9 +1076,9 @@ pango_core_text_font_key_equal (const PangoCoreTextFontKey *key_a,
 }
 
 int
-pango_core_text_font_key_get_absolute_size (const PangoCoreTextFontKey *key)
+pango_core_text_font_key_get_size (const PangoCoreTextFontKey *key)
 {
-  return key->pixelsize;
+  return key->pointsize;
 }
 
 double
@@ -1152,32 +1165,34 @@ pango_core_text_font_map_new_font (PangoCoreTextFontMap    *fontmap,
 static gboolean
 find_best_match (PangoCoreTextFamily         *font_family,
                  const PangoFontDescription  *description,
-                 PangoFontDescription       **best_description,
                  PangoCoreTextFace          **best_face)
 {
   PangoFontDescription *new_desc;
+  PangoFontDescription *best_description = NULL;
   int i;
 
-  *best_description = NULL;
   *best_face = NULL;
 
   for (i = 0; i < font_family->n_faces; i++)
     {
       new_desc = pango_font_face_describe (font_family->faces[i]);
 
-      if (pango_font_description_better_match (description, *best_description, new_desc))
+      if (pango_font_description_better_match (description, best_description,
+                                               new_desc))
 	{
-	  pango_font_description_free (*best_description);
-	  *best_description = new_desc;
+	  pango_font_description_free (best_description);
+	  best_description = new_desc;
 	  *best_face = (PangoCoreTextFace *)font_family->faces[i];
 	}
       else
 	pango_font_description_free (new_desc);
     }
 
-  if (*best_description)
-    return TRUE;
-
+  if (best_description)
+    {
+      pango_font_description_free (best_description);
+      return TRUE;
+    }
   return FALSE;
 }
 
@@ -1199,6 +1214,16 @@ pango_core_text_font_map_get_serial (PangoFontMap *fontmap)
   return ctfontmap->serial;
 }
 
+static void
+pango_core_text_font_map_changed (PangoFontMap *fontmap)
+{
+  PangoCoreTextFontMap *ctfontmap = PANGO_CORE_TEXT_FONT_MAP (fontmap);
+
+  ctfontmap->serial++;
+  if (ctfontmap->serial == 0)
+    ctfontmap->serial++;
+}
+
 static PangoFont *
 pango_core_text_font_map_load_font (PangoFontMap               *fontmap,
                                     PangoContext               *context,
@@ -1471,6 +1496,7 @@ pango_core_text_font_map_class_init (PangoCoreTextFontMapClass *class)
   fontmap_class->load_fontset = pango_core_text_font_map_load_fontset;
   fontmap_class->shape_engine_type = PANGO_RENDER_TYPE_CORE_TEXT;
   fontmap_class->get_serial = pango_core_text_font_map_get_serial;
+  fontmap_class->changed = pango_core_text_font_map_changed;
 }
 
 /*
@@ -1511,10 +1537,12 @@ G_DEFINE_TYPE (PangoCoreTextFontset,
                pango_core_text_fontset,
                PANGO_TYPE_FONTSET);
 
+#if !defined(MAC_OS_X_VERSION_10_8) || MAC_OS_X_VERSION_MIN_REQUIRED < MAC_OS_X_VERSION_10_8
 /* This symbol does exist in the CoreText library shipped with Snow
  * Leopard and Lion, however, it is not found in the public header files.
  */
 CFArrayRef CTFontCopyDefaultCascadeList (CTFontRef font_ref);
+#endif
 
 static PangoCoreTextFontset *
 pango_core_text_fontset_new (PangoCoreTextFontsetKey    *key,
@@ -1523,66 +1551,97 @@ pango_core_text_fontset_new (PangoCoreTextFontsetKey    *key,
   PangoCoreTextFamily *font_family;
   PangoCoreTextFontset *fontset;
   PangoCoreTextFont *best_font = NULL;
+  gchar **family_names;
   const gchar *family;
   gchar *name;
+  GPtrArray *fonts;
+  int i;
 
+  fonts = g_ptr_array_new ();
   family = pango_font_description_get_family (description);
-  family = family ? family : "";
-  name = g_utf8_casefold (family, -1);
-  font_family = g_hash_table_lookup (key->fontmap->families, name);
-  g_free (name);
+  family_names = g_strsplit (family ? family : "", ",", -1);
 
-  if (font_family)
+  for (i = 0; family_names[i]; ++i)
     {
-      PangoFontDescription *best_description;
-      PangoCoreTextFace *best_face;
-      gint size;
-      gboolean is_absolute;
+      name = g_utf8_casefold (family_names[i], -1);
+      font_family = g_hash_table_lookup (key->fontmap->families, name);
+      g_free (name);
 
-      /* Force a listing of the available faces */
-      pango_font_family_list_faces ((PangoFontFamily *)font_family, NULL, NULL);
+      if (font_family)
+        {
+          PangoCoreTextFace *family_face;
+          PangoCoreTextFont *font;
 
-      if (!find_best_match (font_family, description, &best_description, &best_face))
-	return NULL;
+          /* Force a listing of the available faces */
+          pango_font_family_list_faces ((PangoFontFamily *)font_family, NULL, NULL);
 
-      size = pango_font_description_get_size (description);
-      if (size < 0)
-        return NULL;
+          if (find_best_match (font_family, description, &family_face))
+            {
+              font = pango_core_text_font_map_new_font (key->fontmap,
+                                                        key,
+                                                        family_face->ctfontdescriptor,
+                                                        family_face->synthetic_italic);
 
-      is_absolute = pango_font_description_get_size_is_absolute (description);
-      if (is_absolute)
-        pango_font_description_set_absolute_size (best_description, size);
-      else
-        pango_font_description_set_size (best_description, size);
-
-      best_font = pango_core_text_font_map_new_font (key->fontmap,
-                                                     key,
-                                                     best_face->ctfontdescriptor,
-                                                     best_face->synthetic_italic);
-
-      pango_font_description_free (best_description);
+              if (font)
+                {
+                  g_ptr_array_add (fonts, font);
+                  if (best_font == NULL) best_font = font;
+                }
+            }
+        }
     }
-  else
-    return NULL;
+
+  g_strfreev (family_names);
 
   if (!best_font)
-    return NULL;
+    {
+      g_ptr_array_free (fonts, false);
+      return NULL;
+    }
 
   /* Create a font set with best font */
   fontset = g_object_new (PANGO_TYPE_CORE_TEXT_FONTSET, NULL);
   fontset->key = pango_core_text_fontset_key_copy (key);
   fontset->orig_description = pango_font_description_copy (description);
 
-  fontset->fonts = g_ptr_array_new ();
-  g_ptr_array_add (fontset->fonts, best_font);
+  fontset->fonts = fonts;
   fontset->coverages = g_ptr_array_new ();
 
   /* Add the cascade list for this language */
+#if defined(MAC_OS_X_VERSION_10_8) && MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_8
+    {
+      CFArrayRef language_pref_list = NULL;
+      CFStringRef languages[1];
+
+      if (key->language)
+        {
+          languages[0] = CFStringCreateWithCString (NULL,
+                                                    pango_language_to_string (key->language),
+                                                    kCFStringEncodingASCII);
+          language_pref_list = CFArrayCreate (kCFAllocatorDefault,
+                                              (const void **) languages,
+                                              1,
+                                              &kCFTypeArrayCallBacks);
+        }
+
+      fontset->cascade_list = CTFontCopyDefaultCascadeListForLanguages (pango_core_text_font_get_ctfont (best_font), language_pref_list);
+
+      if (language_pref_list)
+        {
+          CFRelease (languages[0]);
+          CFRelease (language_pref_list);
+        }
+    }
+#else
+  /* There is unfortunately no public API to retrieve the cascade list
+   * on Mac OS X < 10.8, so we use the following undocumented public function.
+   */
   fontset->cascade_list = CTFontCopyDefaultCascadeList (pango_core_text_font_get_ctfont (best_font));
+#endif
 
   /* length of cascade list + 1 for the "real" font at the front */
-  g_ptr_array_set_size (fontset->fonts, CFArrayGetCount (fontset->cascade_list) + 1);
-  g_ptr_array_set_size (fontset->coverages, CFArrayGetCount (fontset->cascade_list) + 1);
+  g_ptr_array_set_size (fontset->fonts, CFArrayGetCount (fontset->cascade_list) + fonts->len);
+  g_ptr_array_set_size (fontset->coverages, CFArrayGetCount (fontset->cascade_list) + fonts->len);
 
   return fontset;
 }
